#include "Environment.h"
#include "GvsPreprocessor.h"
#include "GlRenderer.h"
#include "VssRay.h"
#include "ViewCellsManager.h"
#include "Triangle3.h"
#include "IntersectableWrapper.h"
#include "Plane3.h"
#include "RayCaster.h"
#include "Exporter.h"


namespace GtpVisibilityPreprocessor
{
  
struct VizStruct
{
	Polygon3 *enlargedTriangle;
	Triangle3 originalTriangle;
	VssRay *ray;
};

static vector<VizStruct> vizContainer;

GvsPreprocessor::GvsPreprocessor(): Preprocessor(), mSamplingType(0), mCastPassSamples(0), mCastTotalSamples(0)
{
	Environment::GetSingleton()->GetIntValue("GvsPreprocessor.totalSamples", mTotalSamples);
	Environment::GetSingleton()->GetIntValue("GvsPreprocessor.initialSamples", mInitialSamples);
	Environment::GetSingleton()->GetIntValue("GvsPreprocessor.samplesPerPass", mSamplesPerPass);
	Environment::GetSingleton()->GetFloatValue("GvsPreprocessor.epsilon", mEps);
	Environment::GetSingleton()->GetFloatValue("GvsPreprocessor.threshold", mThreshold);	

	Debug << "Gvs preprocessor options" << endl;
	Debug << "number of total samples: " << mTotalSamples << endl;
	Debug << "number of initial samples: " << mInitialSamples << endl;
	Debug << "number of samples per pass: " << mSamplesPerPass << endl;
	Debug << "threshold: " << mThreshold << endl;
	Debug << "eps: " << mEps << endl;

	mStats.open("gvspreprocessor.log");
}


bool GvsPreprocessor::CheckDiscontinuity(const VssRay &currentRay,
										 const Triangle3 &hitTriangle,
										 const VssRay &oldRay)
{
	const float dist = Magnitude(oldRay.GetDir());
	const float newDist = Magnitude(currentRay.GetDir());
 
#if 0
	if ((dist - newDist) > mThresHold)
#else // rather take relative distance
	if ((dist / newDist) > mThreshold)
#endif
	{
		VssRay *newRay = ReverseSampling(currentRay, hitTriangle, oldRay);
		if (!HandleRay(newRay))
		{
			delete newRay;
		}

		return true;
	}

	return false;
}


bool GvsPreprocessor::HandleRay(VssRay *vssRay)
{
	mViewCellsManager->ComputeSampleContribution(*vssRay, true, false);
		
	// some pvs contribution for this ray?
	if (vssRay->mPvsContribution > 0)
	{
		cout << " h " << mCastPassSamples << " " << mSamplesPerPass << " " << mTotalSamples << " q: " << mRayQueue.size() << endl;
		
		cout << "error!!! " << vssRay->mPvsContribution << " " << vssRay->mTerminationObject << endl;

		mRayQueue.push(vssRay);
		mVssRays.push_back(new VssRay(*vssRay));
		++ mCastPassSamples;
		return true;
	}

	return false;
}


/** Creates 3 new vertices for triangle vertex with specified index.
*/
static void CreateNewVertices(VertexContainer &vertices,
							  const Triangle3 &hitTriangle,
							  const VssRay &ray, 
							  const int index,
							  const float eps)
{
	const int indexU = (index + 1) % 3;
	const int indexL = (index == 0) ? 2 : index - 1;

	const Vector3 a = hitTriangle.mVertices[index] - ray.GetOrigin();
	const Vector3 b = hitTriangle.mVertices[indexU] - hitTriangle.mVertices[index];
	const Vector3 c = hitTriangle.mVertices[index] - hitTriangle.mVertices[indexL];
	
	const float len = 1;//Magnitude(a);

	const Vector3 dir1 = Normalize(CrossProd(a, b)); //N((pi-xp)(pi+1- pi));
	const Vector3 dir2 = Normalize(CrossProd(a, c)); // N((pi-xp)(pi- pi-1))
	const Vector3 dir3 = DotProd(dir2, dir1) > 0 ? // N((pi-xp)di,i-1+di,i+1(pi-xp))
		Normalize(dir2 + dir1) : Normalize(CrossProd(a, dir1) + CrossProd(dir2, a)); 

	// compute the new three hit points
	// pi, i + 1:  pi+ e|pi-xp|di, j
	const Vector3 pt1 = hitTriangle.mVertices[index] + eps * len * dir1;
	// pi, i - 1:  pi+ e|pi-xp|di, j
    const Vector3 pt2 = hitTriangle.mVertices[index] + eps * len * dir2;
	// pi, i:  pi+ e|pi-xp|di, j
	const Vector3 pt3 = hitTriangle.mVertices[index] + eps * len * dir3;
	
	vertices.push_back(pt2);
	vertices.push_back(pt3);
	vertices.push_back(pt1);
}


void GvsPreprocessor::EnlargeTriangle(VertexContainer &vertices,
									  const Triangle3 &hitTriangle,
									  const VssRay &ray)
{
	CreateNewVertices(vertices, hitTriangle, ray, 0, mEps);
	CreateNewVertices(vertices, hitTriangle, ray, 1, mEps);
	CreateNewVertices(vertices, hitTriangle, ray, 2, mEps);
}


static Vector3 CalcPredictedHitPoint(const VssRay &newRay, 
									 const Triangle3 &hitTriangle,
									 const VssRay &oldRay)
{
	Plane3 plane(hitTriangle.GetNormal(), hitTriangle.mVertices[0]);

	const Vector3 hitPt = 
		plane.FindIntersection(newRay.mTermination, newRay.mOrigin);
	
	return hitPt;
}


static bool EqualVisibility(const VssRay &a, const VssRay &b)
{
	return a.mTerminationObject == b.mTerminationObject;
}


int GvsPreprocessor::SubdivideEdge(const Triangle3 &hitTriangle,
								   const Vector3 &p1, 
								   const Vector3 &p2, 
								   const VssRay &x, 
								   const VssRay &y,
								   const VssRay &oldRay)
{
	// the predicted hitpoint expects to hit the same mesh again
	const Vector3 predictedHitX = CalcPredictedHitPoint(x, hitTriangle, oldRay);
	const Vector3 predictedHitY = CalcPredictedHitPoint(y, hitTriangle, oldRay);

	CheckDiscontinuity(x, hitTriangle, oldRay);
	CheckDiscontinuity(y, hitTriangle, oldRay);

	if (EqualVisibility(x, y))
	{
		return 2;
	}
	else
	{
		const Vector3 p = (p1 + p2) * 0.5f;
		SimpleRay sray(oldRay.mOrigin, p - oldRay.mOrigin);
	
		VssRay *newRay = mRayCaster->CastSingleRay(sray, mViewSpaceBox);

		if (!newRay) return 0;
		const bool enqueued = HandleRay(newRay);
		
		const int s1 = SubdivideEdge(hitTriangle, p1, p, x, *newRay, oldRay);
		const int s2 = SubdivideEdge(hitTriangle, p, p2, *newRay, y, oldRay);
		return s1 + s2;
		
		if (!enqueued)
			delete newRay;
	}
}


int GvsPreprocessor::AdaptiveBorderSampling(const VssRay &currentRay)
{
	cout << "a";
	Intersectable *tObj = currentRay.mTerminationObject;
	Triangle3 hitTriangle;

	// other types not implemented yet
	if (tObj->Type() == Intersectable::TRIANGLE_INTERSECTABLE)
	{
		hitTriangle = dynamic_cast<TriangleIntersectable *>(tObj)->GetItem();
	}
	else
	{
		cout << "not yet implemented" << endl;
	}

	VertexContainer enlargedTriangle;
	
	/// create 3 new hit points for each vertex
	EnlargeTriangle(enlargedTriangle, hitTriangle, currentRay);
	
	/// create rays from sample points and handle them
	SimpleRayContainer simpleRays;
	simpleRays.reserve(9);

	VertexContainer::const_iterator vit, vit_end = enlargedTriangle.end();

	for (vit = enlargedTriangle.begin(); vit != vit_end; ++ vit)
	{
		const Vector3 rayDir = (*vit) - currentRay.GetOrigin();
		SimpleRay sr(currentRay.GetOrigin(), rayDir);
		simpleRays.AddRay(sr);

		//cout << "here244 " << (*vit) << " " << currentRay.GetOrigin() << endl;
		//cout << " sr" << sr.mOrigin << sr.mDirection << endl;
	}

	VizStruct dummy;
	dummy.enlargedTriangle = new Polygon3(enlargedTriangle);
	dummy.originalTriangle = hitTriangle;
	//dummy.ray = new VssRay(currentRay);
	vizContainer.push_back(dummy);

	// cast rays to triangle vertices and determine visibility
	VssRayContainer vssRays;
	CastRays(simpleRays, vssRays, false, false);
	//CastRays(simpleRays, vssRays, true, true);
	//cout << "here211 " << simpleRays.size() << " " << vssRays.size() << endl;
	// add to ray queue
	EnqueueRays(vssRays);
	
	const int n = (int)enlargedTriangle.size();
	
#if 0
    // recursivly subdivide each edge
	for (int i = 0; i < n; ++ i)
	{
		SubdivideEdge(
			hitTriangle,
			enlargedTriangle[i], 
			enlargedTriangle[(i + 1) % n], 
			*vssRays[i], 
			*vssRays[(i + 1) % n],
			currentRay);
	}
#endif
	return (int)vssRays.size();
}


static Vector3 GetPassingPoint(const VssRay &currentRay,
										 const Triangle3 &hitTriangle,
										 const VssRay &oldRay)
{
	// intersect triangle plane with plane spanned by current samples
	Plane3 plane(currentRay.GetOrigin(), currentRay.GetTermination(), oldRay.GetTermination());
	Plane3 triPlane(hitTriangle.GetNormal(), hitTriangle.mVertices[0]);

	SimpleRay intersectLine = GetPlaneIntersection(plane, triPlane);

	// Evaluate new hitpoint just outside the triangle
	const float factor = 0.95f;
	float t = triPlane.FindT(intersectLine);

	const Vector3 newPoint = intersectLine.mOrigin + t * factor * intersectLine.mDirection;

	return newPoint;
}


VssRay *GvsPreprocessor::ReverseSampling(const VssRay &currentRay,
										 const Triangle3 &hitTriangle,
										 const VssRay &oldRay)
{
	cout << "r" << endl;
	//-- The plane p = (xp, hit(x), hit(xold)) is intersected
	//-- with the newly found triangle (xold is the previous ray from
	//-- which x was generated). On the intersecting line, we select a point
	//-- pnew which lies just outside of the new triangle so the ray
	//-- just passes by inside the gap
    const Vector3 newPoint = GetPassingPoint(currentRay, hitTriangle, oldRay);
	const Vector3 predicted = CalcPredictedHitPoint(currentRay, hitTriangle, oldRay);

	//-- Construct the mutated ray with xnew,dir = predicted(x)- pnew 
	//-- as direction vector
	const Vector3 newDir = predicted - newPoint ;
	// take xnew,p = intersect(viewcell, line(pnew, predicted(x)) as origin ?
	// difficult to say!!
	const Vector3 newOrigin = newDir * -5000.0f;

	return new VssRay(currentRay);
}


int GvsPreprocessor::CastInitialSamples(const int numSamples, 
										const int sampleType)
{	
	const long startTime = GetTime();

	// generate simple rays
	SimpleRayContainer simpleRays;
	GenerateRays(numSamples, sampleType, simpleRays);

	//cout << "here909 " << simpleRays.size() << endl;
	//for (int i = 0; i < simpleRays.size(); ++ i)
	//	cout << "here544 " << simpleRays[i].mOrigin << endl;
	// generate vss rays
	VssRayContainer samples;
	CastRays(simpleRays, samples, true);
	
	// add to ray queue
	EnqueueRays(samples);
//for (int i = 0; i < samples.size(); ++ i)
//		cout << "here1544 " << samples[i]->mOrigin << endl;
	Debug << "generated " <<  numSamples << " samples in " << TimeDiff(startTime, GetTime()) * 1e-3 << " secs" << endl;

	return (int)samples.size();
}


void GvsPreprocessor::EnqueueRays(VssRayContainer &samples)
{
	// add samples to ray queue
	VssRayContainer::const_iterator vit, vit_end = samples.end();
	
	for (vit = samples.begin(); vit != vit_end; ++ vit)
	{
		HandleRay(*vit);
	}
}


int GvsPreprocessor::Pass()
{
	const int mSampleType = 0;
	// reset samples
	mCastPassSamples = 0;
	
	while (mCastPassSamples < mSamplesPerPass) 
	{	
		// Ray queue empty => 
		// cast a number of uniform samples to fill ray Queue
		CastInitialSamples(mInitialSamples, mSampleType);
		ProcessQueue();

		cout << "\nround finished, cast " << mCastPassSamples << " of " << mSamplesPerPass << endl;
	}

	mCastTotalSamples += mCastPassSamples;

	return mCastPassSamples;
}


int GvsPreprocessor::ProcessQueue()
{
	int castSamples = 0;

	while (!mRayQueue.empty()) 
	{
		// handle next ray
		VssRay *ray = mRayQueue.top();
		mRayQueue.pop();
	
		castSamples += AdaptiveBorderSampling(*ray);
		delete ray;
	}
	
	return castSamples;
}


bool GvsPreprocessor::ComputeVisibility()
{
	Randomize(0);
	const long startTime = GetTime();
	
	mViewSpaceBox = mKdTree->GetBox();
	cout << "Gvs Preprocessor started\n" << flush;
	
	if (!mLoadViewCells)
	{	/// construct the view cells from the scratch
		ConstructViewCells(mViewSpaceBox);
		cout << "view cells loaded" << endl;
	}

	while (mCastTotalSamples < mTotalSamples) 
	{
		Pass();
				
		////////
		//-- stats
		cout << "+";
		cout << "\nsamples cast " << mCastPassSamples << " (=" << mCastTotalSamples << " of " << mTotalSamples << ")" << endl;
		//mVssRays.PrintStatistics(mStats);
		mStats << "#Time\n" << TimeDiff(startTime, GetTime())*1e-3 << endl
			   << "#TotalSamples\n" << mCastTotalSamples << endl;

		mViewCellsManager->PrintPvsStatistics(mStats);
		// ComputeRenderError();
	}

	Visualize();	
	return true;
}


void GvsPreprocessor::Visualize()
{
	Exporter *exporter = Exporter::GetExporter("gvs.wrl");

	if (!exporter)
		return;
	
	vector<VizStruct>::const_iterator vit, vit_end = vizContainer.end();
	for (vit = vizContainer.begin(); vit != vit_end; ++ vit)
	{
		exporter->SetWireframe();
		exporter->ExportPolygon((*vit).enlargedTriangle);
		//Material m;
		exporter->SetFilled();
		Polygon3 poly = Polygon3((*vit).originalTriangle);
		exporter->ExportPolygon(&poly);
	}

	exporter->ExportRays(mVssRays);
	delete exporter;
}

}